#  Copyright (c) 2025, Greg Michael
#  Licensed under BSD 3-Clause License. See LICENSE.txt for details.
#
#  Adapted from geopandas/plotting.py
#  Copyright (c) 2013-2016, GeoPandas developers.

import numpy as np
from matplotlib.path import Path
from matplotlib.patches import PathPatch
from matplotlib.collections import PatchCollection
def shp_plot_polygon(ax, poly, **kwargs):

    path = Path.make_compound_path(
        Path(np.asarray(poly.exterior.coords)[:, :2]),
        *[Path(np.asarray(ring.coords)[:, :2]) for ring in poly.interiors])
    patch = PathPatch(path, **kwargs)
    collection = PatchCollection([patch], **kwargs)

    ax.add_collection(collection, autolim=True)


#  Copyright (c) 2021-2025, Greg Michael
#  Licensed under BSD 3-Clause License. See LICENSE.txt for details.

import os
import sys
import argparse
import re
import platform
import subprocess
import shlex

import craterstats as cst
import craterstats.gm as gm


class AppendPlotDict(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        s=' '.join(values)
        d = {}
        for kv in re.split(r',(?=\w+=)',s): # only split on commas directly preceding keys
            try:
                k, v = kv.split("=")
            except:
                sys.exit("Missing '=' in expression: " + option_string+' '+kv)
            d[k] = v
        list_of_d = getattr(namespace, self.dest)
        list_of_d=[d] if list_of_d is None else list_of_d+[d]
        setattr(namespace, self.dest, list_of_d)

class SpacedString(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, ' '.join(values))

class LoadFromFile(argparse.Action):
    def __call__ (self, parser, namespace, values, option_string = None):
        try:
            with open(values) as f:
                # parse arguments in the file and store them in the target namespace
                a=f.read()
                a=('\n').join([e for e in a.split('\n') if e.strip() and not e.strip().startswith('#')] ) #remove '#' commented lines
                parser.parse_args(shlex.split(a), namespace)
                namespace.input_filename = f.name
                setattr(namespace, self.dest, True)
        except FileNotFoundError:
            raise argparse.ArgumentError(self, f"File not found: {values}")
        except PermissionError:
            raise argparse.ArgumentError(self, f"Permission denied: {values}")
        except Exception as e:
            raise argparse.ArgumentError(self, e)

def get_parser():
    parser = argparse.ArgumentParser(description='Craterstats: a tool to analyse and plot crater count data for planetary surface dating.')

    parser.add_argument('-i','--input', help="input args from file", metavar='filename.cs', action=LoadFromFile)
    parser.add_argument('-inc', '--include', help="include settings from external file", metavar='filename.txt', action=LoadFromFile)

    parser.add_argument("-lcs", help="list chronology systems", action='store_true')
    parser.add_argument("-lpc", help="list plot symbols and colours", action='store_true')
    parser.add_argument("-about", help="show program details", action='store_true')
    parser.add_argument("-v","--version", help="show program version", action='store_true')
    parser.add_argument("-demo", help="run sequence of demonstration commands: output in ./demo", action='store_true')
    parser.add_argument("-b","--bins", help="show bin boundaries", action='store_true')
    parser.add_argument("--convert", help="desired format [stat|scc|shp], source file", nargs=2, type=str)

    parser.add_argument("-o","--out", help="output filename (omit extension for default) or directory", nargs='+', action=SpacedString)
    parser.add_argument("--functions_user", help="path to file containing user defined chronology systems", nargs='+', action=SpacedString)
    parser.add_argument("--create_desktop_icon", help="create desktop icon for activated window", action='store_true')
    parser.add_argument("-m", "--merge", help="merge crater count files", nargs='+', action=SpacedString)

    parser.add_argument("-f", "--format", help="output formats",  nargs='+', choices=['png','tif','pdf','svg','csv'])

    parser.add_argument("-cs", "--chronology_system", help="chronology system index")
    parser.add_argument("-ef", "--equilibrium", help="equilibrium function index")
    parser.add_argument("-ep", "--epochs", help="epoch system index")

    parser.add_argument("-title", help="plot title", nargs='+', action=SpacedString)
    parser.add_argument("-pr", "--presentation", help="data presentation: " + (', ').join(cst.PRESENTATIONS))
    parser.add_argument("-xrange", help="x-axis range, log(min) log(max)", nargs=2)
    parser.add_argument("-yrange", help="y-axis range, log(min) log(max)", nargs=2)
    parser.add_argument("-isochrons", help="comma-separated isochron list in Ga, e.g. 1,3,3.7a,4a (optional combined suffix to modify label: n - suppress; a - above; s - small)")
    parser.add_argument("-legend", help="0 - suppress; or any combination of: n - name, a - area, p - perimeter, c - number of craters, r - range, N - N(d_ref) value")
    parser.add_argument("-mu", nargs='?', choices=[0,1], type=int, const=1, help="1 - show; 0 - suppress")
    parser.add_argument("-style", choices=['natural', 'root-2'], help="diameter axis style")

    parser.add_argument("-invert",  nargs='?', choices=[0,1], type=int, const=1, help="1 - invert to black background; 0 - white background")
    parser.add_argument("-text_halo", nargs='?', choices=[0,1], type=int, const=1, help="1 - on [default]; 0 - off")
    parser.add_argument("-transparent", nargs='?', choices=[0,1], type=int, const=1, help="set transparent background")
    #combine invert/transparent into one? maybe not, but invert could be same syntax - get rid of 0,1
    parser.add_argument("-tight", help="tight layout", action='store_true')

    parser.add_argument("-pd", "--print_dimensions", help="print dimensions: either single value (cm/decade) or enclosing box in cm (AxB), e.g. 2 or 8x8")
    parser.add_argument("-pt_size", type=float, help="point size for figure text")
    parser.add_argument("-ref_diameter", type=float, help="reference diameter for displayed N(d_ref) values")
    parser.add_argument("-sf","--sig_figs", type=int, choices=[2,3], help="number of significant figures for displayed ages")

    parser.add_argument("-st","--sequence_table", help="generate sequence probability table", action='store_true')

    parser.add_argument("-d_min","--min_diameter", type=float, help="minimum diameter for uncertainty plot")
    parser.add_argument("-ns", "--n_samples", type=int, help="number of samples for uncertainty plot")

    parser.add_argument("-p", "--plot", nargs='+', action=AppendPlotDict, metavar="KEY=VAL,",
                        help="Specify overplot.\nAllowed keys:   \n"
                             "source=txt,"
                             "name=txt,"
                             "range=[min,max],"
                             "snap={1,0},"
                             "type={data,poisson,b-poisson,c-fit,d-fit},"
                             "error_bars={1,0},"
                             "hide={1,0},"
                             "colour={0-31},"
                             "psym={0-14},"
                             "binning={" +','.join(cst.Cratercount.BINNINGS) + "},"
                             "age_left={1,0},"
                             "show_age={1,0},"
                             "resurf={1,0}, apply resurfacing correction,"
                             "resurf_showall={1,0}, show all data with resurfacing correction,"
                             "isochron={1,0}, show whole fitted isochron,"
                             "offset_age=[x,y], in 1/20ths of decade")

    parser.add_argument("-ra", "--randomness_analysis", help="source fil for randomness analysis", nargs='+', action=SpacedString)
    parser.add_argument("-trials", type=int, help="number of Monte Carlo trials for randomness analysis")
    parser.add_argument("-measure", help="comma-separated list of measures for randomness analysis (from m2cnd,sdaa)")
    parser.add_argument("-ra_offset", type=int, help="vertical offset for randomness analysis sub-plot in 1/20ths of decade")

    return parser

def defaults():
    set = {
        'chronology_system': 'Moon, Neukum (1983)',
        'cite_functions': 1,
        'epochs': None,
        'equilibrium': None,
        'invert': 0,
        'transparent':0,
        'text_halo':1,
        'isochrons': None,
        'legend': 'fnacr',
        'mu': 1,
        'presentation': 'differential',
        'print_dimensions': '7.5x7.5',
        'pt_size': 9.0,
        'randomness': 0,
        'ref_diam': 1,
        'sig_figs': 3,
        'style': 'natural',
        'title': None,
        'format': {'png', 'csv'},
        'min_diameter':0.15,
        'global_area':1e12, # default larger than all terrestrial planets
        'n_samples':200,
        'ra_offset':0,
        }
    plot = {
        'source': '',
        'name': '',
        'range': ['0', 'inf'],
        'snap': 1,
        'type': 'data',
        'error_bars': 1,
        'hide': 0,
        'colour': 0,
        'psym': 1,
        'binning': 'pseudo-log',
        'age_left': 0,
        'show_age': 1,
        'resurf': 0,
        'resurf_showall': 0,
        'isochron': 0,
        'offset_age': [0, 0]
    }
    return {'set':set,'plot':plot}


def decode_abbreviation(s,v,one_based=False,allow_ambiguous=False,allow_invalid=False):
    """
    decode arbitrary abbreviation of list member into index

    :param s: full string list
    :param v: abbreviation or index
    :param one_based: menu indexing of s starts at 1?
    :param allow_ambiguous: if allowed, return first match
    :return: index (zero-based)
    """

    #perhaps add back index selection requiring # prefix?

    regex = '(?i)' + '.*'.join(v)
    res = [(i,e) for i,e in enumerate(s) if re.search(regex, e) is not None]
    res.sort(key=lambda x: len(x[1]))
    if len(res) == 0:
        if allow_invalid:
            return -1
        else:
            sys.exit('Invalid abbreviation: ' + v)
    elif len(res) > 1 and not allow_ambiguous:
        sys.exit('Ambiguous abbreviation: ' + v)
    return res[0][0]


def construct_cps_dict(args,c,f,default_filename):
    if 'presentation' in vars(args):
        if args.presentation is not None:
            c['presentation'] = cst.PRESENTATIONS[decode_abbreviation(cst.PRESENTATIONS, args.presentation,one_based=True)]
    if c['presentation'] in ['chronology', 'rate', 'sequence','uncertainty']: #possible to overwrite with user-choice
        c['xrange'] = cst.DEFAULT_XRANGE[c['presentation']]
        c['yrange'] = cst.DEFAULT_YRANGE[c['presentation']]
    if c['presentation']=='sequence':
        c['legend']='A'

    for k,v in vars(args).items():
        if v is None:
            if k == 'out': c[k] = default_filename # don't set as default in parse_args: need to detect None in source_cmds
        else:
            if k in ('title',
                     'isochrons',
                     'legend',
                     'print_dimensions',
                     'pt_size',
                     'ref_diameter',
                     'sig_figs',
                     'randomness',
                     'mu',
                     'invert',
                     'transparent',
                     'text_halo',
                     'style',
                     'xrange', 'yrange',
                     'min_diameter','n_samples',
                     'bins',
                     'ra_offset',
                     ):
                c[k]=v
            elif k in ('chronology_system','equilibrium','epochs'):
                names = [e['name'] for e in f[k]]
                c[k]=f[k][decode_abbreviation(names, v, one_based=True, allow_ambiguous=True)]['name']

            elif k == 'out':
                if os.path.isdir(v):
                    c[k] = os.path.normpath(v+'/'+default_filename)
                else:
                    c[k] = gm.filename(v, 'pn')
                    ext = gm.filename(v, 'e').lstrip('.')
                    if ext: c['format'].add(ext)
            elif k == 'format':
                c[k]=set(v)


    cs=next((e for e in f['chronology_system'] if e['name'] == c['chronology_system']), None)
    if cs is None: sys.exit('Chronology system not found:' + c['chronology_system'])

    c['cf'] = cst.Chronologyfn(f, cs['cf'])
    c['pf'] = cst.Productionfn(f, cs['pf'])
    i=decode_abbreviation(cst.PLANETS,cs['body'],allow_invalid=True)
    if i!=-1:
        c['global_area']=cst.SURFACE_AREAS[i]

    if 'equilibrium' in c and c['equilibrium'] not in (None,''):
        c['ef'] = cst.Productionfn(f, c['equilibrium'], equilibrium=True)
    if 'epochs' in c and c['epochs'] not in (None,''):
        c['ep'] = cst.Epochs(f, c['epochs'],c['pf'],c['cf'])

    if c['presentation'] == 'Hartmann':
        if hasattr(c['pf'],'xrange'): #not possible to overwrite with user choice
            c['xrange'] = c['pf'].xrange
            c['yrange'] = c['pf'].yrange
        else:
            c['xrange'] = cst.DEFAULT_XRANGE['Hartmann']
            c['yrange'] = cst.DEFAULT_YRANGE['Hartmann']

    return c


def construct_plot_dicts(args,plot):
    #if type(plot) is list: plot=plot[0] #take only first plot entry as template
    cpl = []
    specified_source = False
    if args.plot is None: return []
    for d in args.plot:
        p=plot.copy()
        if cpl: # for these items: if not given, carry over from previous
            for k in ['source','psym','snap','isochron','error_bars','colour','binning']:
                p[k] = cpl[-1][k]
            if p['source'] == cpl[-1]['source']: #only carry type if source unchanged
                p['type'] = cpl[-1]['type']

        for k0,v in d.items():
            k=cst.CRATERPLOT_KEYS[decode_abbreviation(cst.CRATERPLOT_KEYS, k0, allow_ambiguous=True)]
            if k in (
                    'name',
                    'snap',
                    'error_bars',
                    'hide',
                    'age_left',
                    'show_age',
                    'resurf',
                    'resurf_showall',
                    'isochron',
                    ):
                p[k]=v
            elif k in ('range','offset_age'):
                p[k] = v.strip('[]').split(',')
            elif k == 'source':
                p['source']=cs_source(v).strip('"')
                specified_source = True
            elif k == 'type':
                p[k]=cst.OPLOT_TYPES_SHORT[decode_abbreviation(cst.OPLOT_TYPES, v, allow_ambiguous=True)]
            elif k == 'binning':
                p[k] = cst.Cratercount.BINNINGS[decode_abbreviation(cst.Cratercount.BINNINGS, v, allow_ambiguous=True)]
                cst.Cratercount.BINNINGS
            elif k == 'colour':
                if v[0]=='#':
                    if bool(re.match(r'^#([0-9A-Fa-f]{6})$', v)):
                        p[k]=v
                    else:
                        sys.exit('Invalid colour:'+v)
                else:
                    names=[n for c1,c2,n in cst.PALETTE]
                    p[k]=decode_abbreviation(names, v, allow_ambiguous=True)
            elif k == 'psym':
                j=[i for i,e in enumerate(cst.MARKERS) if e[0]==v]
                if len(j)==1: # found standard abbreviation
                    p[k]=j[0]
                else: # look for arbitrary abbreviation or index
                    names = [e[1] for e in cst.MARKERS]
                    p[k]=decode_abbreviation(names, v, allow_ambiguous=True)

        if not specified_source: sys.exit('Source not specified')
        if os.path.isabs(p['source']):
            src = p['source']
        else:
            src = (gm.filename(args.input_filename,'p') if args.input else '')+p['source']
        p['cratercount'] = cst.Cratercount(src)
        cpl += [p]
    return cpl

def convert_format(args, cps, cs_content):
    def outfile(name, out, ext):
        if out:
            if os.path.isdir(out):
                outfile = out + name + ext
            else:
                outfile = gm.filename(out, 'pn') + ext
        else:
            outfile = name + ext
        return outfile

    fmt, src = args.convert
    scc = cst.Spatialcount(src.replace('%sample%/', cst.PATH + 'sample/'))
    fmt1 = fmt.lstrip(".")
    fmt0 = gm.filename(src,'e').lstrip(".")
    out = outfile(gm.filename(src, 'n'), args.out, '.'+fmt1)
    match fmt1:
        case 'stat' if fmt0 in ['diam','scc','shp']:
            cc = cst.Cratercount(src)
            out = gm.filename(out, 'p1e', cc.name)
            cc.WriteStatFile(out)
        case 'scc':
            if fmt0 == 'shp':
                scc = cst.Spatialcount(src)
                out = gm.filename(out, 'p1e', scc.name)
                scc.writeSCCfile(out)
        case 'shp':
            if fmt0 == 'scc':
                scc = cst.Spatialcount(src)
                scc.writeSHPfiles(out)
                out = gm.filename(out,'pn1e', '[_CRATER,_AREA]')
        case 'png'|'svg'|'pdf':
            if fmt0 in ['shp','scc']:
                out = re.sub(r'_?CRATER_?', '', out)
                out = gm.filename(out, 'pn1e','_map')
                cps.create_map_plotspace()
                scc.plot(cps,grid=True)
                cps.fig.savefig(out, dpi=500, transparent=cps.transparent, bbox_inches='tight' if args.tight else None, pad_inches=.02 if args.tight else None)
                gm.write_textfile(gm.filename(out,'pn1','.cs'), cs_content)
        case _:
            print(f"{fmt0} to {fmt1} conversion not supported")
            return
    print(f"Conversion written to: {out}")
    return

def source_cmds(src):
    cmd=gm.read_textfile(src, ignore_blank=True, ignore_hash=True)
    parser=get_parser()
    for i,c in enumerate(cmd):
        a = c.split()
        args = parser.parse_args(a)
        if args.out is None:
            f=f'{i:02d}-out'
            a+=['-o',f]
            c+=' -o '+f
        print(f'\nCommand: {i}\ncraterstats ' + c)
        main(a)
    print('\nProcessing complete.')

def demo(d=None,src=None):
    if not src: src = cst.PATH+'config/demo_commands.txt'
    cmd = gm.read_textfile(src, ignore_blank=True, ignore_hash=True)
    out='demo/'
    os.makedirs(out,exist_ok=True)
    if d is None:
        d=range(0,len(cmd))
    for i in d:
        c=cmd[i]
        print(f'\nDemo {i}\ncraterstats '+c)
        a = shlex.split(f'-o {out}{i:02d}-demo {c}')
        main(a)
    print('\nDemo output written to: '+out)
    return d

def do_functions_user(args):
    if os.environ.get('CONDA_PREFIX'):
        config = os.environ.get('CONDA_PREFIX')+'/etc/config_functions_user.txt'
    else:
        return None #actions environment is not conda
    if args.functions_user:
        s=['#path to functions_user.txt',args.functions_user]
        gm.write_textfile(config,s)
        print(': '.join(s))
    return gm.read_textfile(config,ignore_hash=True)[0] if gm.file_exists(config) else None

def create_desktop_icon():
    system = platform.system()
    match system:
        case 'Windows':
            subprocess.run([
                "powershell", "-NoProfile", "-Command", '''
                $s = New-Object -ComObject WScript.Shell
                $sc = $s.CreateShortcut("$env:USERPROFILE\\Desktop\\craterstats-III.lnk")
                $sc.TargetPath = "$env:windir\\system32\\cmd.exe"
                $sc.Arguments = '/K "%USERPROFILE%\\miniforge3\\Scripts\\activate.bat craterstats"'
                $sc.WorkingDirectory = "$env:USERPROFILE"
                $sc.Save()
                '''
            ], check=True)
            print('Desktop shortcut created.')

        case _:
            print(f"Desktop shortcut creation is not yet implemented for: {system}.")

def cs_source(v):
    return v.replace('%sample%/', cst.PATH + 'sample/')

def randomness_analysis(args,cps):
    out = '' if cps.out=='out' else gm.filename(cps.out,'pn')
    ra = cst.Randomnessanalysis(cs_source(args.randomness_analysis), out=out)
    cps.out = gm.filename(ra.ra_file,'pn')
    trials = args.trials if args.trials else 300
    cps.measures = args.measure.split(',') if args.measure else ['m2cnd','sdaa']
    diff = set(cps.measures) - {'m2cnd','sdaa'}
    if diff:
        sys.exit(f"Invalid measure: {diff}")
    for measure in cps.measures:
        ra.run_montecarlo(trials, measure)
        # do each loop so as to retain data if interrupted
        ra.calculate_stats()
        ra.write()
    return ra

def main(args0=None):
    args = get_parser().parse_args(args0)
    if not args0: args0=sys.argv[1:]

    functions=cst.PATH+'config/functions.txt'
    s = gm.read_textfile(functions, ignore_hash=True, strip=';', as_string=True)  # read and remove comments
    functions_user = do_functions_user(args)
    if functions_user:
        try:
            s = s + gm.read_textfile(functions_user, ignore_hash=True, strip=';', as_string=True)
        except:
            print("Unable to read user functions file: "+functions_user+" - ignoring.")
    functions = gm.read_textstructure(s,from_string=True)

    if args.create_desktop_icon:
        create_desktop_icon()
        return

    if args.lcs:
        print(gm.bright("\nChronology systems:"))
        print('\n'.join([f'{e['name']}' for e in functions['chronology_system']]))
        print(gm.bright("\nEquilibrium functions:"))
        print('\n'.join([f'{e['name']}' for e in functions['equilibrium']]))
        print(gm.bright("\nEpoch systems:"))
        print('\n'.join([f'{e['name']}' for e in functions['epochs']]))
        return

    if args.lpc:
        print(gm.bright("\nPlot symbols:"))
        print(', '.join([f'{e[1]} ({e[0]})' for e in cst.MARKERS]))
        print(gm.bright("\nColours:"))
        print(', '.join([f'{e[2]}' for e in cst.PALETTE]))
        return

    if args.about:
        print('\n'.join(cst.ABOUT))
        return

    if args.version:
        print('\n'.join(cst.ABOUT[0:5]))
        return

    if args.merge:
        cst.merge_cratercounts(args)
        return

    if args.demo:
        demo()
        return

    dflt = defaults()
    cp_dicts = construct_plot_dicts(args,dflt['plot'])
    if args.input:
        default_filename = gm.filename(args.input_filename,'pn')
    else:
        default_filename = '_'.join(sorted(set([gm.filename(d['source'], 'n') for d in cp_dicts]))) if cp_dicts else 'out'
        default_filename = re.sub(r'_?CRATER_?', '', default_filename) # remove if present from shp file
    cps_dict = construct_cps_dict(args, dflt['set'], functions, default_filename)
    if gm.filename(cps_dict['out'],'n')=='out' and cps_dict['presentation'] in ('chronology', 'rate', 'uncertainty'):
        cps_dict['out']=gm.filename(cps_dict['out'],'p')+cps_dict['presentation']

    if 'a' in cps_dict['legend'] and 'b-poisson' in [d['type'] for d in cp_dicts]:
        cps_dict['legend']+='p' #force to show perimeter with area if using b-poisson

    cps=cst.Craterplotset(cps_dict) #,craterplot=cpl)
    for d in cp_dicts:
        if isinstance(d['colour'], int):d['colour']=cps.palette[d['colour']]
    cpl = [cst.Craterplot(d) for d in cp_dicts]
    cps.craterplot=cpl

    cs_content = ''.join(['\n'+e if e[0]=='-' and not (e+' ')[1].isdigit() else ' '+shlex.quote(e) for e in args0])[1:]

    if args.convert:
        convert_format(args, cps, cs_content)
        return

    if cpl and cps.presentation not in ('sequence','uncertainty'):
        cps.autoscale(cps_dict['xrange'] if 'xrange' in cps_dict else None,
                      cps_dict['yrange'] if 'yrange' in cps_dict else None)


    def savefig(tag=''):
        cps.fig.savefig(cps.out + tag +'.' + f, dpi=500, transparent=cps.transparent,
                        bbox_inches='tight' if args.tight else None, pad_inches=.02 if args.tight else None)

    drawn=False
    for f in cps.format:
        if f in {'png','pdf','svg','tif'}:
            if args.randomness_analysis:
                ra = randomness_analysis(args,cps)
                for measure in cps.measures:
                    ra.plot_montecarlo_split(cps, measure)
                    savefig('-'+measure)
                # cps.create_map_plotspace()
                # ra.plot(cps,grid=True)
                # savefig('-map')
            elif cps.presentation == 'uncertainty':
                for plt in ('k','err','age'):
                    cps.draw()
                    cps.age_area_plot(plt)
                    savefig('_'+plt)
            else:
                if not drawn:
                    cps.draw()
                    drawn = True
                savefig()

        if f in {'csv'}:
            cps.create_summary_table(f_out=cps.out+'.'+f)

    if not args.input:
        gm.write_textfile(cps.out + '.cs', cs_content)

if __name__ == '__main__': 
    main()




#  Copyright (c) 2021-2025, Greg Michael
#  Licensed under BSD 3-Clause License. See LICENSE.txt for details.

import numpy as np
from scipy.special import factorial

from . import normal


def poisson(k,lam,cumulative=False,threshold=23):
    '''
    poisson mass function
    vectorised in lambda, but not k

    :param k:
    :param lam:
    :param cumulative: return cmf instead of pmf
    :param threshold: use normal approximation above threshold value of k
    :return: pmf
    '''

    if k<threshold:       # poisson
        if cumulative:
            res = 0.
            for i in range(k+1):
                res += (lam**i) / factorial(i)
            return np.exp(-lam) * res
        else:
            return lam**k * np.exp(-lam) / factorial(k)

    else:                 # normal
        if cumulative:
            return normal(lam, np.sqrt(lam), k, cumulative=True)
        else:
            return normal(lam, np.sqrt(lam), k)







#  Copyright (c) 2025, Greg Michael
#  Licensed under BSD 3-Clause License. See LICENSE.txt for details.

from collections import namedtuple
from concurrent.futures import ProcessPoolExecutor, as_completed
import math
import os

import astropy_healpix as hpx
import astropy.units as u
import numpy as np
import matplotlib.ticker as ticker
from scipy.spatial import SphericalVoronoi
import shapely as shp
import shapely.affinity as shp_aff
import spherely as sph
from progressbar import ProgressBar
import pyproj as prj

import craterstats as cst
import craterstats.gm as gm


class Randomnessanalysis(cst.Spatialcount):
    '''Applies randomness tests to Spatialcount'''

    MEASURES = ['m2cnd','sdaa']
    def __init__(self,filename=None,area_file=None,out=None):
        super().__init__(filename,area_file)
        self.init_Cratercount()
        self.montecarlo = {}
        self.max_threads = os.cpu_count()-1
        self.ra_file = (out if out else self.name) + "_ra.txt"
        self.read()
        binning='root-2'
        self.cc.apply_binning(binning, offset=0.)
        self.plot_reduction_factor = None

    def init_Cratercount(self):
        self.cc = cst.Cratercount()
        self.cc.diam = self.diam
        self.cc.lon = self.lon
        self.cc.lat = self.lat
        self.cc.fraction = self.fraction
        self.cc.prebinned = 0

    def establish_hpx(self,n):
        """
        establish HEALpix grid for 2d binning
        """
        area_globe = 4 * math.pi * self.planetary_radius**2
        desired_nbins = n/10
        desired_bin_area = self.area / desired_nbins
        desired_n_hpx = area_globe / desired_bin_area
        nside = 2**round(np.log2(math.sqrt(desired_n_hpx/12)))
        self.hp = hpx.HEALPix(nside=nside, order='nested')

    def oplot_neighbours(self,cps,craters,neighbours,ax=None):
        """
        Overplot neighbours
        """
        if not ax:
            ax=cps.ax
        sz_ratio = ax.get_position().width/cps.ax.get_position().width
        for lon,lat,nb in zip(craters.lon,craters.lat,neighbours):
            x0, y0 = self.ortho_proj(lon, lat)
            x1, y1 = self.ortho_proj(sph.get_x(nb), sph.get_y(nb))
            ax.plot((x0,x1),(y0,y1), color=cps.grey[0], linewidth=0.5*cps.sz_ratio*sz_ratio, zorder=1)

    def oplot_voronoi(self,cps,craters,p2,ax=None):
        """
        Overplot voronoi polygons
        """
        if not ax:
            ax=cps.ax
        sz_ratio = ax.get_position().width/cps.ax.get_position().width

        # start from sph_polygons output (p2) from sdaa calc.
        shp_polygons0 = [shp.from_wkb(sph.to_wkb(p)) for p in p2]
        # flatten any multipolygons (disjoint areas or holes should also be outlined)
        shp_polygons = [polygon for shape in shp_polygons0 for polygon in (shape.geoms if isinstance(shape, shp.MultiPolygon) else [shape])]

        ns = 20
        frac = np.linspace(0, 1-1/ns, ns)
        geod = prj.Geod(a=1,f=0)

        # need to interpolate between vertices here to get proper curvature in distorted map (only for display)
        # could make ns fn of distance if needed
        interp_threshold = result = 2 * math.pi / 1800
        for p in shp_polygons:
            x0, y0 = p.exterior.xy
            x1,y1 = [],[]
            for i in range(len(x0)):
                i1 = (i + 1) % len(x0)
                azimuth, _, distance = geod.inv(x0[i], y0[i], x0[i1], y0[i1])
                if distance > interp_threshold:
                    x1i, y1i, _ = geod.fwd(np.full(ns, x0[i]), np.full(ns, y0[i]), np.full(ns, azimuth), distance * frac)
                    x1.extend(x1i)
                    y1.extend(y1i)
                else:
                    x1.append(x0[i])
                    y1.append(y0[i])

            x2, y2  = self.ortho_proj(x1, y1)
            projected_polygon = shp.Polygon(zip(x2, y2)) # Reassemble projected polygon
            gm.shp_plot_polygon(ax, projected_polygon, facecolor='none', edgecolor=cps.grey[0], linewidth=0.5 * sz_ratio)


    def get_bin_craters(self,bin,Craterlist=False):
        # from real config
        bin_keys = [f"{e:.3g}" for e in np.log2(self.cc.binned['d_min'])]
        if Craterlist:
            xydf = [(x, y, d, f) for x, y, d, f, i in zip(self.lon, self.lat, self.diam, self. fraction, self.cc.reverse_indices)
                    if bin_keys[i - 1] == bin]
            craters = cst.Spatialcount.Craterlist(lon=[x for x, _, _, _ in xydf],
                                 lat=[y for _, y, _, _ in xydf],
                                 diam=[d for _, _, d, _ in xydf],
                                 fraction=[f for _, _, _, f in xydf])
            return craters
        else:
            xy = [(e, f) for e, f, i in zip(self.lon, self.lat, self.cc.reverse_indices) if bin_keys[i - 1] == bin]
            pts, ids = zip(*[(sph.create_point(longitude=x, latitude=y), self.hp.lonlat_to_healpix(x * u.deg, y * u.deg)) for x, y in xy])
            hpd = {}
            for pt, id in zip(pts, ids):
                if id not in hpd:
                    hpd[id] = [pt]
                else:
                    hpd[id].append(pt)
            return pts,ids,hpd


    def montecarlo_split(self, measure, self_pp, staggered=False):
        """
        Prepare separate runs across bin range
        """
        match measure:
            case 'sdaa': min_count = 3
            case 'm2cnd': min_count = 3

        self.montecarlo[measure]['trials'] = {}

        for b,n in zip(self.cc.binned['d_min'],self.cc.binned['n_event']):
            if n >= min_count:
                bin = f"{np.log2(b):.3g}"
                print(f"{measure}, bin {bin}: {gm.diameter_range([b,b*math.sqrt(2)],2)}, {n} craters")

                # do parallel monte carlo for random configs
                m = montecarlo_pp(self_pp,b,n)
                self.montecarlo[measure]['trials'][bin] = m


    def run_montecarlo(self, trials, measure):

        self.establish_hpx(trials)
        self_pp = self.self_pp(trials, measure)

        if not measure in self.montecarlo or not self.montecarlo[measure]['n_trials'] == trials: # skip montecarlo if already have data
            self.montecarlo[measure] = {'n_trials':trials}
            self.montecarlo_split(measure, self_pp)

    def calculate_stats(self):
        for measure in self.montecarlo.keys():
            self_pp = self.self_pp(self.montecarlo[measure]['n_trials'], measure)
            self.montecarlo[measure]['stats'] = {}
            stats = namedtuple('stats', ['m0', 'p2', 'mn', 'sd', 'n_sigma','percentile'])
            for bin in self.montecarlo[measure]['trials'].keys():

                pts, ids, hpd = self.get_bin_craters(bin)
                m0,p2 = evaluate_randomness(self_pp,pts, ids, hpd)
                #print(f"Bin: {bin} d_min:{2**float(bin):0.3g} number:{len(pts)} Actual value of measure: {m0:0.3g}")

                m = self.montecarlo[measure]['trials'][bin]
                mn,sd = (np.mean(m),np.std(m))
                n_sigma = (m0-mn)/sd
                percentile = np.searchsorted(np.sort(m), m0) / len(m) * 100
                self.montecarlo[measure]['stats'][bin] = stats(m0=m0, p2=p2, mn=mn, sd=sd, n_sigma=n_sigma,percentile = percentile)

    def plot_histogram(self,cps,measure,bin,ax0=None,sz_ratio=1.): # mark median and 1 sd band
        if ax0:
            ax=ax0
        else:
            cps.create_map_plotspace()
            ax = cps.ax

        res = self.montecarlo[measure]['stats'][bin]
        m = self.montecarlo[measure]['trials'][bin]
        m0 = res.m0

        nbins = round(math.sqrt(self.montecarlo[measure]['n_trials']) + 5)

        h,be = np.histogram(m, bins=nbins)
        bar_width = np.diff(be)
        xr = [min(list(be)+[m0])-bar_width[0],max(list(be)+[m0])+bar_width[0]]
        if not ax0:
            xr[0] -= .6*gm.mag(xr)
        yr = [0,max(h)*(1.1 if ax0 else 1.5)]

        ax.bar(be[:-1], h, width=bar_width, color = cps.grey[3], edgecolor=None, linewidth=0.5,align='edge')

        h = list(h)
        x0,y0 = zip(*[(e,f[i]) for e,f in zip(be,zip([0]+h,h+[0])) for i in [0,1]])
        ax.plot(x0, y0, color=cps.palette[0], linewidth=.7 * sz_ratio)

        ax.plot([m0]*2, yr, color=cps.palette[0],alpha = 0.4, lw = 2 * cps.sz_ratio * sz_ratio)

        ax.text(m0, yr[0]+1*gm.mag(yr), f"{m0:0.3g}\n{gm.percentile_sigfigs(res.percentile)}%", size=.7*cps.scaled_pt_size * math.sqrt(sz_ratio), rotation=0,
                                 horizontalalignment='center', verticalalignment='bottom')

        # shade sd
        sdr = [res.mn-res.sd,res.mn+res.sd]
        x1,y1 = zip(*[(e,f) for e,f in zip(x0,y0) if sdr[0]<e<sdr[1]])
        x1 = [sdr[0]]*2+list(x1)+[sdr[1]]*2
        y1 = [0,y1[0]]+list(y1)+[y1[-1],0]
        ax.fill_between(x1, y1, color=cps.grey[1], lw=.5 * cps.sz_ratio * sz_ratio)

        # plot mean
        ax.plot([res.mn] * 2, [0,np.interp(res.mn, x1, y1)], color=cps.palette[0], alpha=0.4, lw=1 * cps.sz_ratio * sz_ratio)

        ax.set_xlim(xr[0], xr[1])
        ax.set_ylim(yr[0], yr[1])

        xtickv = [e for e in gm.ticks(xr, 5) if xr[0] < e < xr[1]]
        xt_label = [f'{e:g}' for e in xtickv]
        xt_label[-1] = "  "*len(xt_label[-1])+xt_label[-1]+(" km" if measure=='m2cnd' else " km²")
        ax.set_xticks(xtickv)
        ax.tick_params(axis='x', which='both', width=.5*sz_ratio, length=cps.pt_size * .3 *sz_ratio, pad=cps.pt_size * .1)
        ax.tick_params(axis='x', which='minor', length=cps.pt_size * .15 * sz_ratio)
        ax.xaxis.set_minor_locator(ticker.AutoMinorLocator())
        ax.set_xticklabels(xt_label,fontsize=.7*cps.scaled_pt_size * math.sqrt(sz_ratio), color=cps.palette[0])

        ax.patch.set_facecolor('none')

        if ax0:
            for s in ['left', 'right', 'top']: ax.spines[s].set_visible(False)
            ax.set_yticks([])
        else:
            ax.set_ylabel('Frequency')

    def get_plot_reduction_factor(self,hf):
        """
        Find reduction factor so that map doesn't overlap histogram
        hf : histogram fractional width of plot
        return: amount to reduce normalised map plot
        """
        if not self.plot_reduction_factor:
            # this could be moved to init....
            cen = sph.centroid(self.polygon)
            cenlon, cenlat = (sph.get_x(cen),sph.get_y(cen))
            ortho_proj = prj.Proj(proj='ortho', lat_0=cenlat, lon_0=cenlon, R=self.planetary_radius)
            self.ortho_proj = ortho_proj

            # do area
            z = sph.to_wkb(self.polygon)
            multipolygon = shp.from_wkb(z)
            p = []

            if not isinstance(multipolygon, shp.MultiPolygon):
                multipolygon = shp.MultiPolygon([multipolygon])

            for poly in multipolygon.geoms:
                exterior_x, exterior_y = poly.exterior.xy
                x_exterior, y_exterior = ortho_proj(exterior_x, exterior_y)
                # Reassemble the projected polygon
                projected_polygon = shp.Polygon(zip(x_exterior, y_exterior))
                p.append(projected_polygon)

            mp = shp.MultiPolygon(p)
            x_mn,y_mn,x_mx,y_mx = shp.bounds(mp)
            x_mg,y_mg = (x_mx-x_mn,y_mx-y_mn)
            mg = max(x_mg,y_mg)
            mp_scaled = shp_aff.scale(mp,xfact=1/mg,yfact=1/mg)
            mp_normalised = shp_aff.translate(mp_scaled,xoff=-mp_scaled.bounds[0], yoff=-mp_scaled.bounds[1])
            hf = .3
            hist_box = shp.Polygon([(1,0),(1,hf*.5),(1-hf,hf*.5),(1-hf,0),(1,0)])
            f = 1.
            while shp_aff.scale(mp_normalised,xfact=f,yfact=f).overlaps(hist_box):
                f *= .98
            self.plot_reduction_factor = f
        return self.plot_reduction_factor

    def plot_map_and_histogram(self, cps, measure, bin, ax=None,sz_ratio=1.):
        if not ax:
            ax=cps.ax
        ax.tick_params(left=False, bottom=False, labelleft=False, labelbottom=False)
        for spine in ax.spines.values(): spine.set_visible(False)
        ax.set_facecolor('none')

        hf = .3
        mf = self.get_plot_reduction_factor(hf)
        offset = (0, .15)
        pos = ax.get_position()
        dx,dy = ( (1-hf) * pos.width * offset[0], pos.height * (1- hf*.5) * offset[1] )
        ax1 = cps.fig.add_axes([pos.x0, pos.y0 + (1-mf) * pos.height, pos.width * mf, pos.height * mf])
        ax2 = cps.fig.add_axes([pos.x0 + (1-hf) * pos.width - dx, pos.y0 +dy, pos.width * hf - dx, pos.height * hf *.5])

        craters = self.get_bin_craters(bin, Craterlist=True)
        self.plot(cps, craters=craters, ax=ax1)
        match measure:
            case 'm2cnd': self.oplot_neighbours(cps,craters,self.montecarlo[measure]['stats'][bin].p2,ax=ax1)
            case 'sdaa': self.oplot_voronoi(cps, craters, self.montecarlo[measure]['stats'][bin].p2, ax=ax1)
        self.plot_histogram(cps, measure, bin, ax0=ax2,sz_ratio=sz_ratio)

        d_min = 2**float(bin)
        ax.text(1., .95, gm.diameter_range([d_min,d_min*2**.499],2)+f'\nn = {len(craters.diam)}', size=.7 * cps.scaled_pt_size * math.sqrt(sz_ratio),
                 transform=ax.transAxes, ha="right",va='top')


    def plot_montecarlo_split(self,cps,measure): # multiplot?
        dim = math.ceil(math.sqrt(len(self.montecarlo[measure]['stats'])+1))
        cps.create_map_plotspace()
        pos = cps.ax.get_position()
        #cps.ax.set_visible(False)
        cps.ax.set_axis_off()
        def make_ax(i):
            x,y = (i%dim,dim - 1 - i//dim)
            axi = cps.fig.add_axes([
                pos.x0 + x * (pos.width / dim),
                pos.y0 + y * (pos.height / dim),  # pos.y0 + (n - 1 - i) * (pos.height / n),
                pos.width / dim,
                pos.height / dim
            ])
            return axi

        for i,bin in enumerate(self.montecarlo[measure]['stats']):
            self.plot_map_and_histogram(cps,measure,bin ,ax=make_ax(i),sz_ratio=1/dim)

        self.plot_n_sigma(cps, measure, ax0=make_ax(dim ** 2 - 1))


    def plot_n_sigma(self, cps, measure, ax0=None):
        if not ax0:
            ax0=cps.ax
        ax0.set_visible(False)
        pos = ax0.get_position()
        sz_ratio = pos.width / cps.ax.get_position().width

        if ax0:
            ax = cps.fig.add_axes([pos.x0+.15*pos.width,pos.y0 + pos.height*.3,pos.width*.95,pos.height*.4])
        else:
            ax = cps.fig.add_axes([pos.x0, pos.y0 + pos.height / 3, pos.width, pos.height / 2])

        x = [np.log10(2**(float(r)+.25)) for r in self.montecarlo[measure]['stats']]
        y0 = [(e.m0 - e.mn)/e.sd for e in self.montecarlo[measure]['stats'].values()]
        y1 = [cst.n_sigma_scaling(e) for e in y0] # apply axis scaling
        y = [-e if measure=='sdaa' else e for e in y1] # flip for sdaa
        xr = min(x)-np.log10(2**.251),max(x)+np.log10(2**.251) # control end-labels
        mg = gm.mag(cps.xrange)

        for spine in set(ax.spines.values())-{'bottom'}: spine.set_visible(False)
        #ax.xaxis.set_visible(False)
        ax.yaxis.set_visible(False)
        ax.set_ylim(cst.n_sigma_scaling(-6), cst.n_sigma_scaling(6))
        ax.set_xlim(xr[0], xr[1])

        ax.fill_between(xr, [cst.n_sigma_scaling(-3)]*2, y2 = [cst.n_sigma_scaling(3)]*2, color=cps.grey[2], edgecolor='none')
        ax.fill_between(xr, [cst.n_sigma_scaling(-2)]*2, y2 = [cst.n_sigma_scaling(2)]*2, color=cps.grey[3], edgecolor='none')
        ax.fill_between(xr, [cst.n_sigma_scaling(-1)]*2, y2 = [cst.n_sigma_scaling(1)]*2, color=cps.grey[1], edgecolor='none')

        marker = cps.marker_def[10].copy()
        marker['markersize'] *= sz_ratio

        ax.plot(x, y, color=cps.palette[0], lw=1. * cps.sz_ratio*sz_ratio, **marker, linestyle=cst.LINESTYLES[measure], clip_on=False)

        dy=-.08
        ax.text(np.mean(xr), cst.n_sigma_scaling(-5), "clustered", color=cps.grey[0], size=.4 * cps.scaled_pt_size, va='center', ha='center', clip_on=False)
        ax.text(np.mean(xr), cst.n_sigma_scaling(7), "separated", color=cps.grey[0], size=.4 * cps.scaled_pt_size, va='center', ha='center', clip_on=False)
        ax.text(xr[1], dy, r"    $n_\sigma$", color=cps.grey[0], size=.4 * cps.scaled_pt_size, va='center', ha='left')
        for y in [-3,-1,0,1,3]:
            ax.text(xr[1], cst.n_sigma_scaling(y)+dy, f"{abs(y):>2}", color=cps.grey[0], size=.3 * cps.scaled_pt_size, va='center', ha='left')

        ax.text(xr[0] - mg*.005, 0, measure + f"\n{self.montecarlo[measure]['n_trials']} trials", color=cps.grey[0], size=.4 * cps.scaled_pt_size,  va='center', ha='right')

        xtickv,xtickname,_,xminorv = cst.Hartmann_bins(xr)
        ax.spines['top'].set_position(('data', cst.n_sigma_scaling(3)))
        ax.xaxis.set_ticks_position('top')
        ax.tick_params(axis='x', which='both', direction="out", color=cps.grey[1], labelcolor=cps.grey[0], width=.5*sz_ratio,
                       length=cps.pt_size*sz_ratio * .2, pad=cps.pt_size*sz_ratio * .2, labelsize=cps.scaled_pt_size*sz_ratio)
        ax.tick_params(axis='x', which='minor', length=cps.pt_size*sz_ratio * .15)
        ax.set_xticks(xtickv)
        ax.set_xticks(xminorv, minor=True)
        ax.set_xticklabels(xtickname)



    def write(self):
        s = ['# Randomness analysis',
              f'version = {cst.__version__}',
              f'source = "{self.filename}"',
              ]

        table = []
        for bin in self.montecarlo[next(iter(self.montecarlo))]['stats'].keys():
            row = f"{bin:<12}"
            for measure in self.montecarlo:
                row += f"\t{self.montecarlo[measure]['stats'][bin].n_sigma:9.4g}"
            table.append(row)
        n_sigma = (['#','n_sigma = {bin' + ''.join([f', {measure}' for measure in self.montecarlo])]
                   + table + ['}'])
        s += n_sigma

        for measure in self.montecarlo:
            trials = (
                ['trials  = {index, ' + ", ".join([b for b in self.montecarlo[measure]['trials']])]
                + ["\t".join([f"{t:<12}"] + [f"{v[t]:<12.7g}" for bin,v in self.montecarlo[measure]['trials'].items()])
                            for t in range(self.montecarlo[measure]['n_trials']) ]
                + ['}'])
            s1 = (
                ['#',f'{measure} = {{',
                f'n_trials = {self.montecarlo[measure]['n_trials']}']
                + trials
                + ['}'])
            s += s1
        gm.write_textfile(self.ra_file,s)

    def read(self):
        # allow read from source file location, but write will still be to current dir or -o
        possible_locations = (self.ra_file, gm.filename(self.filename, 'pn1','_ra.txt'))
        f = next((loc for loc in possible_locations if gm.file_exists(loc)), None)
        if f:
            c = gm.read_textstructure(f)
            for name in self.MEASURES:
                if name in c:
                    self.montecarlo[name] = {
                        'n_trials':int(c[name]['n_trials']),
                        }
                    self.montecarlo[name]['trials'] = {}
                    for b in list(c[name]['trials'].keys())[1:]:
                        self.montecarlo[name]['trials'][b] = [float(e) for e in c[name]['trials'][b]]

    def self_pp(self,trials, measure):
        """
        tuple of vars for parallel processing
        """
        return self_pp_tuple(trials=trials,measure=measure,**{key: getattr(self, key) for key in attrs_pp[:-2]})

#######################################################
# standalone functions required for parallel processing
# note that 'self' in routines below is not class instance, but analogous self_pp namedtuple
#######################################################

attrs_pp = ['hp', 'planetary_radius', 'area', 'enclosing_area', 'polygon', 'xr', 'yr', 'max_threads', 'trials','measure']
self_pp_tuple = namedtuple('self_pp_tuple', attrs_pp)

def evaluate_randomness(self_pp, pts, ids, hpd):
    """
    Evaluate randomness measure
    """
    match self_pp.measure:
        case 'm2cnd':
            measure, p2 = kth_nearest_neighbour_pp(self_pp, pts, ids, hpd, k=2)  # p2 are neighbours (for plotting real config)
        case 'sdaa':
            measure, p2 = sdaa(self_pp, pts, ids, hpd)
    return measure, p2

def run_trial(self_pp, b, n, trial_index):
    pts, ids, hpd = sprinkle_discs_pp(self_pp, n, b)  # Generate points and ids
    m, _ = evaluate_randomness(self_pp, pts, ids, hpd)
    return m

def run_trial_wrapper(args):
    return run_trial(*args)

def montecarlo_pp(self_pp, b, n):
    """
    Single Monte Carlo run. - parallel
    """
    with ProcessPoolExecutor(max_workers=self_pp.max_threads) as executor:
        trial_indices = range(self_pp.trials)
        args = [(self_pp, b, n, trial_index) for trial_index in trial_indices]
        pbar = ProgressBar(max_value=self_pp.trials)

        # Submit tasks to executor and track futures
        future_to_index = {executor.submit(run_trial_wrapper, arg): trial_index for arg, trial_index in zip(args, trial_indices)}

        measures = []
        for future in as_completed(future_to_index):
            result = future.result()  # Get result of completed trial
            measures.append(result)
            pbar.update(future_to_index[future])  # Update progress bar
        pbar.finish()
    return measures

def random_points_pp(self ,n):
    # consider integral of cos(lat): sin(lat) - range varies from -1 to 1 for -180 to 180
    y0 = np.sin(np.radians(self.yr[0])) # move to init?
    y1 = np.sin(np.radians(self.yr[1]))
    y = np.degrees(np.asin(np.random.uniform(y0, y1, n)))
    x = np.random.uniform(self.xr[0], self.xr[1], n)
    return zip(x ,y)

def sprinkle_discs_pp(self ,n ,diam):
    """
    generate z random points in enclosing area
    find those in actual area
    find non-overlapping (could optimise with 2d spatial bin search)
    take first n good points
    """
    expected_hit_rate = self.area /self.enclosing_area
    angular_radius = diam / self.planetary_radius
    hpd = {}
    count = 0

    while count < n:
        shortfall = n - count
        z = int(1.2 * shortfall / expected_hit_rate) + 10  # guess at required number of points
        p0 = random_points_pp(self,z)
        for lon,lat in p0:
            pt = sph.create_point(longitude=lon, latitude=lat)
            if sph.within(pt, self.polygon):

                # erase existing craters closer than diam/2
                hpx_neighbours = self.hp.cone_search_lonlat(lon*u.deg, lat*u.deg, angular_radius*u.rad)
                for id in hpx_neighbours:
                    if id in hpd:
                        len0 = len(hpd[id])
                        hpd[id] = [e for e in hpd[id] if sph.distance(pt, e, radius = self.planetary_radius) > diam/2]
                        count += len(hpd[id]) - len0

                # add new point
                id0 = self.hp.lonlat_to_healpix(lon*u.deg, lat*u.deg)
                if id0 not in hpd:
                    hpd[id0] = [pt]
                else:
                    hpd[id0].append(pt)
                count += 1
                if count >= n: # should never exceed
                    break

    pts,ids = zip(*[(e, key) for key, id_list in hpd.items() for e in id_list])
    return pts,ids,hpd

def kth_nearest_neighbour_pp(self, pts,ids,hpd, k=1):
    """
    find k_th nearest neighbour list
    """
    def get_kth(pt,zone):
        dist_pt = [(sph.distance(pt, e, radius=self.planetary_radius),e)
                   for id in zone if id in hpd
                   for e in hpd[id]]
        sorted_dist_pt = sorted(dist_pt, key=lambda x: x[0])
        return sorted_dist_pt[k]

    neighbours = []
    distances = []

    for id,pt in zip(ids,pts):
        locality = set()
        outer = {id}
        while True:
            locality |= outer # union
            n = sum(len(hpd[e]) for e in locality if e in hpd)
            if n >= k+1:
                break
            hpx_neighbours = set(self.hp.neighbours(tuple(outer)).flatten())
            outer = hpx_neighbours - locality

        d0,pt0 = get_kth(pt,locality)

        # redefine neighbourhood based on known kth closest: if pt near edge, could be over hpx cell boundary
        hpx_neighbours = self.hp.cone_search_lonlat(sph.get_x(pt) * u.deg, sph.get_y(pt) * u.deg, d0/self.planetary_radius * u.rad)
        d1, pt1 = get_kth(pt, hpx_neighbours)
        neighbours += [pt1]
        distances += [d1]

    mean_distance = np.mean(distances)
    return mean_distance,neighbours


def sdaa(self, pts, ids, hpd):
    """
    find standard deviation of adjacent area (spherical)
    """

    ll_radians = [(math.radians(sph.get_x(p)), math.radians(sph.get_y(p))) for p in pts]
    xyz = np.array([(math.cos(y) * math.cos(x), math.cos(y) * math.sin(x), math.sin(y)) for x, y in ll_radians])

    center = np.array([0, 0, 0])
    radius = 1.0
    sv = SphericalVoronoi(xyz, radius, center, threshold=1e-09)
    sv.sort_vertices_of_regions()

    xyz_polygons = [[sv.vertices[id] for id in region] for region in sv.regions]
    ll_polygons = [[(math.degrees(math.atan2(y, x)), math.degrees(math.asin(z))) for x, y, z in p] for p in xyz_polygons]
    sph_polygons0 = [sph.create_polygon([(lon, lat) for lon, lat in p]) for p in ll_polygons]
    sph_polygons = [sph.intersection(self.polygon, p) for p in sph_polygons0]

    areas = [sph.area(p,radius=self.planetary_radius) for p in sph_polygons]
    sdaa = np.std(areas)

    return sdaa, sph_polygons





#  Copyright (c) 2025, Greg Michael
#  Licensed under BSD 3-Clause License. See LICENSE.txt for details.

import numpy as np

def percentile_sigfigs(p, sf=2):
    """
    format string with equivalent differential precision approaching 100 as approaching 0
    p: percentile value between 0 and 100. i.e. 0.015 vs 99.985 as 2 s.f.
    sf: int sig figs
    """

    if p < 50:
        res = f"{p:.{sf}g}"
    else:
        digits = len(f"{100 - p:.{sf}g}")
        res = f"{p:.{max(digits - 2, 0)}f}"
    return res




#  Copyright (c) 2021-2025, Greg Michael
#  Licensed under BSD 3-Clause License. See LICENSE.txt for details.

import sys

import numpy as np
import matplotlib.pyplot as plt

import craterstats as cst
import craterstats.gm as gm


"""
This file contains miscellaneous code for Craterstats
"""

def bin_bias_correction(beta,k):
    '''
    Correction for binning bias of log-log histogram of known slope (see Michael (2013), Eq. 2)

    :param beta: ratio between adjacent bin boundaries, e.g. sqrt(2)
    :param k: log-log slope (including minus sign, e.g. -2)
    :return: correction ratio
    '''

    g=beta**.5-beta**(-.5)
    with np.errstate(invalid='ignore'): # both vectors of conditional are evaluated:
        Fbin2F_ratio = np.where(abs(k) < 1e-5,1., (beta ** (k / 2.) - beta ** (-k / 2.)) / (g * k) )

    return Fbin2F_ratio


class fractional_crater_transform:
    '''
    Transform function between area fraction and linear penetration across boundary.
    Easy to measure the polygon area overlap, but require fractional linear overlap for
    statistics. (Originally suggested by Misha Kreslavsky)

    After initialisation, able to call inverse function af2lf()
    '''

    def __init__(self):
        self.lf = np.array(gm.scl(range(100), out_range=[0., 1]))
        self.af = self.area_fraction(self.lf)

    def area_fraction(self,lf):
        d = 1 - 2 * lf
        theta = 2 * np.arccos(d)
        return (1 / (2 * np.pi)) * (theta - np.sin(theta))

    def af2lf(self,af):
        lf = np.interp(af, self.af, self.lf)
        return lf

    def plot(self):
        plt.plot(self.lf, self.af)
        plt.plot([0,1],[0,1],color='r')
        plt.gca().set_aspect('equal', adjustable='box')
        plt.xlabel('linear frac')
        plt.ylabel('area frac')
        plt.show()


def merge_cratercounts(args,equal_weight=False):
    '''
    Merge two or more cratercounts into a single file

    :param args: cmd line params
    :param: equal_weight [not yet implemented/verified]
    '''
    fs=gm.quoted_split(args.merge)
    if len(fs)<2:
        sys.exit("Specify at least two crater count files.")

    out=args.out
    if not out:
        out='_'.join(sorted(gm.filename(fs, 'n')))
    out = gm.filename(out, 'pn1', '.diam')

    d=[]
    frac=[]
    a=0.
    p=0.

    for f in fs:
        cc=cst.Cratercount(f)
        if cc.prebinned:
            sys.exit("Not a valid crater diameter file: "+f)

        #fraction1 = n_elements(fraction) eq 1?replicate(fraction, n_elements(diam)): fraction
        area=cc.area

        if equal_weight:
            fraction = [e/area for e in cc.fraction]  # weight by area
            area = 1.
        else:
            fraction = cc.fraction

        d += cc.diam
        frac += fraction
        a += area
        p += cc.perimeter

    s = (['# Merged crater diameter file', '#----------------------------', '#']
           + ['# ' + gm.filename(f, "ne") for f in fs]
           + ['#']
           + [f"area = {a}"]
           + [f"perimeter = {p}"]
           + ['#',"crater = {diameter, fraction"]
           + [f"{a:<12g}  {b:g}" for a,b in zip(d,frac)]
           + ['}']
         )

    gm.write_textfile(out,s)
    print('Merged file: '+out)




def str_age(age,e1=0,e2=0,sf=3,unit=None,simple=False,no_error=False,mu=False,MathML=False):
    '''
    Convert numerical age in Ga to latex-style string, optionally with sub/superscript errors and mu uncertainty function

    :param age:
    :param e1: plus error
    :param e2: minus error
    :param sf: no of sig figs
    :param unit: force specified unit
    :param simple: don't show zero decimals (for isochrons), nor use latex-style
    :param no_error: don't show errors
    :param mu: show mu uncertainty function
    :param MathML - use MathML instead of latex. Can paste into MS Word
    :return: age string
    '''

    f = [1, 1e-3, 1e-6, 1e-9]
    unit0 = ['Ga', 'Ma', 'ka', 'a']

    if unit is not None:
        a = next((e, u) for e, u in zip(f, unit0) if unit == u)
    else:
        a = next((e, u) for e, u in zip(f, unit0) if age >= e or u =='a')

    v=age/a[0]

    if simple:
        st=f'{v:g}'
    else:
        st= gm.sigfigs(v, sf)

    if e1>0 and not no_error :
        e1s = gm.sigfigs(e1 / a[0], sf - 1)
        e2s = gm.sigfigs(e2 / a[0], sf - 1)
        if MathML:
            st = '<msubsup><mtext>'+st+'</mtext><mtext>&#x2212;'+e2s+'</mtext><mtext>+'+e1s+'</mtext></msubsup>'
        else:
            st += '^{+' + e1s + '}_{-' + e2s + '}'

    if simple:
        st += ' ' + a[1]
    elif MathML:
        st = ('<math xmlns="http://www.w3.org/1998/Math/MathML">'
              +('<mi>&#956;</mi>' if mu else '')+st+'<mtext>&#8239;'+a[1]+'</mtext></math>')
    else:
        st = '$' + (r'𝜇' if mu else '') + st + '$ ' + a[1]

    return st


def Hartmann_bins(d_range):
    """
    Generate tick values and tick labels with units for Hartmann scale
    """
    v = [np.log10(2. ** e) for e in range(-10, 11)]  # set up root-2/hartmann axis labels
    labels = [str(e) for e in
              (1, 2, 4, 8, 16, 31, 63, 125, 250, 500, 1, 2, 4, 8, 16, 32, 64, 128, 256, 512, 1024)]
    labels[10] = '  ' + labels[10] + "km"
    xtickv, xtickname = map(list, zip(*[(val, txt) for val, txt in zip(v, labels) if d_range[0] <= val <= d_range[1]]))
    if xtickv[0] < 0: xtickname[0] += 'm'
    xminor = (xtickv[1] - xtickv[0]) / 2
    xminorv = [e+xminor for e in [xtickv[0]-2*xminor]+xtickv if d_range[0] <= e+xminor <= d_range[1]]
    return xtickv,xtickname,xminor,xminorv

def n_sigma_scaling(v):
    """
    scaling for n_sigma plot: enlarge central zone
    """
    return np.power(np.abs(v),.6)*np.sign(v)



#  Copyright (c) 2025, Greg Michael
#  Licensed under BSD 3-Clause License. See LICENSE.txt for details.

from collections import namedtuple
from datetime import datetime
import math
import os
import re
import shutil

import numpy as np
import matplotlib.pyplot as plt
import shapefile  # pyshp
import shapely as shp
import spherely as sph
import pyproj as prj

import craterstats as cst
import craterstats.gm as gm

class Spatialcount:
    '''Reads spatial crater count data'''

    Craterlist = namedtuple('Craterlist', ['lon','lat','diam','fraction'])

    def __init__(self,filename=None,area_file=None):
        self.filename=filename
        filetype = gm.filename(filename, 'e') if filename else None
        self.name = gm.filename(filename, "n")

        if filetype == '.scc':
            self.readSCCfile()
        elif filetype == '.shp':
            if not area_file:  # if area_file not specified, try to get from crater_file name
                f = re.sub(r'CRATER', 'AREA', gm.filename(filename, 'n'))
                area_file = gm.filename(filename, 'p1e', f)
            self.crater_file = filename
            self.area_file = area_file
            self.name = re.sub(r'_?CRATER_?', '', gm.filename(filename, "n"))
            self.readSHPfiles()

    def summary(self):
        print(f"\nPlanetary radius: {self.planetary_radius:0g} km")
        print(f"Total area {self.area:.7g} km2\nperimeter: {self.perimeter:.5g} km")
        print("\nDiameter, km  fraction      lon                  lat")
        for d, f, x, y in zip(self.diam, self.fraction, self.lon, self.lat):
            print(f"{d:<12.7g} {f:9.3g} {x:20.12f} {y:20.12f}")

    def llr2xyz(self,lon,lat,r = None):
        if not r: r = self.planetary_radius
        lat_rad = np.radians(lat)
        lon_rad = np.radians(lon)
        x = r * np.cos(lat_rad) * np.cos(lon_rad)
        y = r * np.cos(lat_rad) * np.sin(lon_rad)
        z = r * np.sin(lat_rad)
        return x,y,z

    def readSCCfile(self):

        s = gm.read_textfile(self.filename,ignore_hash=True,strip=';', as_string=True)
        s = re.sub(r"a-axis radius", "oct_a_axis_radius", s) # fix OpenCraterTool misformatting
        c = gm.read_textstructure(s,from_string=True)

        crater=c['crater']
        diam=[float(e) for e in crater['diam']]
        frac=[float(e) for e in crater['fraction']] if 'fraction' in c else [1. for e in diam]
        lon=[float(e) for e in crater['lon']]
        lat = [float(e) for e in crater['lat']]
        q=[i for i,e in sorted(enumerate(diam),key=lambda x:x[1])]     #get sorted indices
        self.diam=[diam[e] for e in q]
        self.fraction=[frac[e] for e in q]
        self.lon = [lon[e] for e in q]
        self.lat = [lat[e] for e in q]

        radius_key = next((k for k in ("a_axis_radius", "oct_a_axis_radius") if k in c), None) # remove later
        self.planetary_radius = float(c[radius_key].split(' ')[0])

        b = c['unit_boundary']
        z = sorted(set([(int(a), b == 'int') for a, b in zip(b['sub_area'], b['tag'])]))

        d={}
        for i, _ in z:
            pts = [(float(x),float(y)) for x,y, sub_area in zip(c['unit_boundary']['lon'],c['unit_boundary']['lat'], c['unit_boundary']['sub_area']) if int(sub_area) == i]
            p = sph.create_polygon(shell=pts)
            d[i] = (pts, p) #, xyz(pts)) # xyz needs to be moved out
            xr1 = gm.range([x for x,y in pts])
            yr1 = gm.range([y for x, y in pts])
            if i == 1:
                xr=xr1
                yr=yr1
            else:
                xr = (min(xr1[0], xr[0]), max(xr1[1], xr[1]))
                yr = (min(yr1[0], yr[0]), max(yr1[1], yr[1]))

        p=[]
        for i, internal in z:
            if not internal:
                holes=[]
                for j, internal1 in z[i:]:
                    if sph.within(d[j][1],d[i][1]):
                        holes += [d[j][0]]
                p += [sph.create_polygon(shell=d[i][0], holes=holes)]

        p = sph.create_multipolygon(p) #merge if multiple

        self.area=sph.area(p, radius=self.planetary_radius)
        self.perimeter=sph.perimeter(p, radius=self.planetary_radius)
        self.polygon=p

        n_pole = sph.create_point(longitude=0, latitude=90)
        s_pole = sph.create_point(longitude=0, latitude=-90)
        if sph.covers(p, n_pole):
            yr = (yr[0], 90)
            xr = (-180,180)
        if sph.covers(p, s_pole):
            yr = (-90, yr[1])
            xr = (-180, 180)
        self.xr=xr
        self.yr=yr
        enclosing_polygon = sph.create_polygon(shell=[(xr[0],yr[0]),(xr[0],yr[1]),(xr[1],yr[1]),(xr[1],yr[0])])
        self.enclosing_area = sph.area(enclosing_polygon, radius=self.planetary_radius)

        print(f"Range: {xr[0]:.2f}:{xr[1]:.2f} {yr[0]:.2f}:{yr[1]:.2f}, area: {self.area:.2f} km2, perimeter: {self.perimeter:.2f} km (planetary radius: {self.planetary_radius:0g} km)")



    def readSHPfiles(self):
        """
        read crater and area shapefiles, extracting diameters, area, perimeter

        """
        def read_shp(shp):
            mtime = os.path.getmtime(shp)
            prj_file = gm.filename(shp, 'pn1', '.prj')
            wkt = gm.read_textfile(prj_file)[0]
            crs = prj.CRS.from_wkt(wkt)
            planetary_radius = crs.ellipsoid.semi_major_metre / 1e3
            sf = shapefile.Reader(shp)
            shapes = sf.shapes()

            if not crs.is_geographic:
                transformer = prj.Transformer.from_crs(crs, crs.geodetic_crs, always_xy=True)
                for shape in shapes:
                    xy = list(zip(*shape.points))
                    lons, lats = transformer.transform(xy[0], xy[1])
                    shape.points = [(lon, lat) for lon, lat in zip(lons, lats)]

            return shapes, planetary_radius, mtime

        sc, rc, tc = read_shp(self.crater_file)
        sa, ra, ta = read_shp(self.area_file)
        if rc != ra:
            raise ValueError("Crater/Area shapefile planetary radii disagree")
        self.planetary_radius = rc
        self.tc, self.ta = tc, ta
        self.sa = sa

        multipolygon = []
        for shape in sa:
            parts = shape.parts
            points = shape.points
            n_rings = len(parts) - 1
            if n_rings == 0:
                p = sph.create_polygon(shell=points)
            else:
                ext = points[:parts[1]]
                int = []
                for i in range(1, len(parts)):
                    start_idx = parts[i]
                    end_idx = parts[i + 1] if i + 1 < len(parts) else len(points)
                    int += [points[start_idx:end_idx]]
                p = sph.create_polygon(shell=ext, holes=int)
            # print(f"Polygon has {n_rings} hole(s). Sub-area: {sph.area(p, radius=ra):.5g} km2")
            multipolygon += [p]
        p = sph.create_multipolygon(multipolygon)
        self.polygon = p
        self.area = sph.area(p, radius=ra)
        self.perimeter = sph.perimeter(p, radius=ra)

        diam = []
        frac = []
        lon = []
        lat = []
        transform = cst.fractional_crater_transform()
        for shape in sc:
            c = sph.create_polygon(shell=shape.points)
            c_area = sph.area(c, radius=rc)
            # diam1 = math.sqrt(4*c_area/math.pi)  # flat diameter from area: better behaviour if noisy vertex positions
            # diam2 = sph.perimeter(c,radius = rc)/math.pi # flat diameter from perimeter: greater error if noisy vertex positions
            diam3 = 2 * rc * math.acos(1 - c_area / (2 * math.pi * rc ** 2))  # spherical diameter from area

            pt = sph.centroid(c)
            x, y = (sph.get_x(pt), sph.get_y(pt))
            # print(f"{len(shape.points)} {(diam2-diam1)/diam1:12.3g} {diam1:12.7g} {diam2:12.7g} {diam3:12.9g} {x:20.12f} {y:20.12f} ")
            diam += [diam3]
            lon += [x]
            lat += [y]
            if sph.covered_by(c, p):
                f = 1
            else:
                area_intersection = sph.area(sph.intersection(c, p), radius=rc)
                area_frac = area_intersection / c_area
                f = transform.af2lf(area_frac)
            frac += [f]

        self.diam = diam
        self.fraction = frac
        self.lon = lon
        self.lat = lat
        return


    def writeSCCfile(self, filename=None):
        sub_area, tag, pts = [],[],[]
        i = 1
        for shape in self.sa: # might be better to use technique from write_SHP, i.e. dismantle self.polygon
            parts = list(shape.parts) + [len(shape.points)]
            pts += shape.points
            for j in range(len(shape.parts)):
                n = parts[j + 1] - parts[j]
                tag += ['ext' if j==0 else 'int'] * n
                sub_area += [i] * n
                i += 1
        boundary_table = [f"{i:<5}\t{a:<5}\t{b:<4}\t{c[0]:23.15f}\t{c[1]:23.15f}"
                          for i, (a, b, c) in enumerate(zip(sub_area, tag, pts), start=1)]

        crater_table = [f"{d:<12.7g}\t{f:9.3g}\t{x:23.15f}\t{y:23.15f}\t{1}"
                        for d, f, x, y in zip(self.diam, self.fraction, self.lon, self.lat)]

        # include so that can import into OpenCratertool (also requires topo_scale_factor for craters - prefer without when n/a)
        s2 = (['#', '# Extra lines for OpenCraterTool compatibility:',
               f"a-axis radius = {self.planetary_radius:0g} <km>", ]
              + [f"# Area_name {i} = NULL" for i in range(1, sub_area[-1] + 1)]
              + ['#'])

        s = (['# Spatial crater count',
              'craterstats_version = '+ cst.__version__,'#',
              '# Derived from:',
              '# ' + datetime.fromtimestamp(self.tc).strftime("%Y-%m-%d %H:%M ") + gm.filename(self.crater_file, 'ne'),
              '# ' + datetime.fromtimestamp(self.ta).strftime("%Y-%m-%d %H:%M ") + gm.filename(self.area_file, 'ne'),
              '#',
              f"a_axis_radius = {self.planetary_radius:0g} <km>",
              f"Total_area = {self.area:0g} <km2>",
              f"Total_perimeter = {self.perimeter:0g} <km>",
              '#',] + s2 + [
              'unit_boundary = {vertex, sub_area, tag, lon, lat']
             + boundary_table +
             ['}',
              'crater = {diam, fraction, lon, lat, topo_scale_factor']
              + crater_table +
              ['}'])

        if filename:
            gm.write_textfile(filename,s)
        else:
            print('\n'.join(s))


    def writeSHPfiles(self, filename):
        """
        Write shapefile pair
        """
        # do crater file
        fc = gm.filename(filename,'pn1e',"_CRATER")
        c = shapefile.Writer(fc)
        c.field('Diam_km', 'F', 12,6)
        c.field('x_coord', 'F', 20,15)
        c.field('y_coord', 'F', 20, 15)
        c.field('tag', 'C', 20)
        rims,wkt = self.find_rims()
        for d, x, y, r in zip(self.diam, self.lon, self.lat, rims):
            c.record(Diam_km=d, x_coord=x, y_coord=y, tag='standard')
            c.poly([list(zip(*r))])
        c.close()
        gm.write_textfile(gm.filename(fc,'pn1','.prj'),wkt)
        shutil.copy(cst.PATH + 'config/_CRATER.qml', gm.filename(fc,'pn1','.qml'))

        # do area file
        fa = gm.filename(filename, 'pn1e', "_AREA")
        a = shapefile.Writer(fa)
        a.field('area', 'F', 12,6)
        a.field('area_name', 'C', 30)
        decomposed = self.polygon_to_pts()
        for i, p in enumerate(decomposed):
            a.poly(p)
            a.record(None, f'Area_{i + 1}')
        a.close()
        gm.write_textfile(gm.filename(fa, 'pn1', '.prj'), wkt)
        shutil.copy(cst.PATH + 'config/_AREA.qml', gm.filename(fa, 'pn1', '.qml'))

    def find_rims(self,craters=None,ns=100):
        """
        Calculate crater rim lon lat points given centres and diameters
        """
        lon,lat,diam = (craters.lon,craters.lat,craters.diam) if craters else (self.lon,self.lat,self.diam)

        theta = [2 * math.pi * i / ns for i in range(ns + 1)]
        cx, cy = np.array([(math.cos(e), math.sin(e)) for e in theta]).T
        rs = f'{self.planetary_radius * 1e3:0.0f}'
        wkt = f'GEOGCS["Spherical_GCS_{rs}", DATUM["Sphere_{rs}", SPHEROID["Sphere_{rs}", {rs}, 0]], PRIMEM["Greenwich", 0], UNIT["Degree", 0.0174532925199433]]'
        geod = prj.CRS(wkt)
        rims = []
        # make projection centre transformers at this spacing for speed (Distance distortion: 0.061% at 2 deg)
        # nb: max offset from projection centre is grid/2
        grid = 5
        t_dict = {}
        for d, x, y in zip(diam, lon, lat):
            origin = (round(x/grid)*grid,round(y/grid)*grid)
            if not origin in t_dict:
                proj4 = f"+proj=laea +lat_ts=0 +lat_0={origin[1]} +lon_0={origin[0]} +R={self.planetary_radius * 1e3:0.0f} +units=m +no_defs"
                laea = prj.CRS(proj4)
                t_dict[origin] = prj.Transformer.from_crs(laea, geod, always_xy=True)  # laea.geodetic_crs
                t_dict[(origin,'inverse')] = prj.Transformer.from_crs(geod, laea, always_xy=True)
            dx,dy = t_dict[(origin,'inverse')].transform(x,y)
            r = d * 1e3 / 2
            ll_pts = t_dict[origin].transform(dx+ cx * r, dy + cy * r)
            rims += [ll_pts]
        return rims,wkt

    def polygon_to_pts(self,p0=None):
        """
        Decompose polygon/multipolygon into list of lists of rings/holes
        """
        if not p0:
            p0=self.polygon
        decomposed=[]
        for p in p0 if isinstance(p0, list) else [p0]:
            z = sph.to_wkb(p)
            y = shp.from_wkb(z)
            x = shp.get_exterior_ring(y)
            rings = [list(x.coords)]
            for j in range(shp.get_num_interior_rings(y)):
                x = shp.get_interior_ring(y,j)
                rings.append(list(x.coords))  # Append holes
            decomposed += [rings]
        return decomposed


    def plot(self,cps,craters=None,grid=False,ax=None):
        """
        Plot Spatialcount

        """
        if not ax:
            ax=cps.ax
        sz_ratio = ax.get_position().width / cps.ax.get_position().width

        cen = sph.centroid(self.polygon)
        cenlon, cenlat = (sph.get_x(cen),sph.get_y(cen))
        xr0 = yr0 = None
        ortho_proj = prj.Proj(proj='ortho', lat_0=cenlat, lon_0=cenlon, R=self.planetary_radius)
        self.ortho_proj = ortho_proj

        # do area
        z = sph.to_wkb(self.polygon)
        multipolygon = shp.from_wkb(z)

        if not isinstance(multipolygon, shp.MultiPolygon):
            multipolygon = shp.MultiPolygon([multipolygon])

        for poly in multipolygon.geoms:
            exterior_x, exterior_y = poly.exterior.xy
            interior_coords = [interior.xy for interior in poly.interiors]
            x_exterior, y_exterior = ortho_proj(exterior_x, exterior_y)
            x_interior = []
            y_interior = []
            for interior_x, interior_y in interior_coords:
                x_interior_tmp, y_interior_tmp = ortho_proj(interior_x, interior_y)
                x_interior.append(x_interior_tmp)
                y_interior.append(y_interior_tmp)

            # Reassemble the projected polygon
            projected_polygon = shp.Polygon(zip(x_exterior, y_exterior),
                holes=[list(zip(xi, yi)) for xi, yi in zip(x_interior, y_interior)] if x_interior else None)
            gm.shp_plot_polygon(ax, projected_polygon, facecolor=cps.grey[2], edgecolor=cps.grey[1], linewidth=0.5*sz_ratio)

            xr0 = gm.range(list(x_exterior) + list(xr0) if xr0 else x_exterior)
            yr0 = gm.range(list(y_exterior) + list(yr0) if yr0 else y_exterior)

        # do craters
        rims,wkt = self.find_rims(craters=craters,ns=30)
        for r in rims:
            x, y = ortho_proj(r[0],r[1])
            ax.plot(x, y, color=cps.palette[0], linewidth=0.3*cps.sz_ratio,zorder=2)

            xr0 = gm.range(list(x) + list(xr0) )
            yr0 = gm.range(list(y) + list(yr0) )

        xr = np.array(gm.range(xr0)) + np.array([-1, 1]) * gm.mag(xr0) * (.1 if ax is cps.ax else .02)
        yr = np.array(gm.range(yr0)) + np.array([-1, 1]) * gm.mag(yr0) * (.1 if ax is cps.ax else .02)

        # do gridlines - this probably needs improving for poles
        if grid:
            lonr, latr = ortho_proj(xr0, yr0, inverse=True)

            xtickv = gm.ticks(lonr, 6)
            ytickv = gm.ticks(latr, 6)

            ns=50
            dlat = ytickv[1]-ytickv[0]
            dlon = xtickv[1]-xtickv[0]
            offset = (ax.transData.transform_point((0, dlat))[1] - ax.transData.transform_point((0, 0))[1])/2e4*cps.scaled_pt_size
            # Plot parallels (latitude lines) and meridians (longitude lines)
            for lat in ytickv[1:-1]:
                x_par, y_par = ortho_proj(np.linspace(xtickv[0]-dlat, xtickv[-1]+dlat, ns), np.repeat(lat,ns))  # plot meridians
                ax.plot(x_par, y_par, color=cps.grey[0], linewidth=0.3*cps.sz_ratio)
                ax.text(np.clip(x_par[-1],xr[0],xr[1]),np.clip(y_par[-1],yr[0],yr[1])-offset,f'{lat:.7g}°',
                        ha='right',va='top',color=cps.grey[0],fontsize=cps.scaled_pt_size*.5)
            for lon in xtickv[1:-1]:
                x_mer, y_mer = ortho_proj( np.repeat(lon,ns), np.linspace(ytickv[0]-dlon, ytickv[-1]+dlon, ns))  # plot parallels
                ax.plot(x_mer, y_mer, color=cps.grey[0], linewidth=0.3*cps.sz_ratio)
                ax.text(np.clip(x_mer[0],xr[0],xr[1]),np.clip(y_mer[0],yr[0],yr[1]),f' {lon:.7g}°',
                        ha='left',va='bottom',color=cps.grey[0],fontsize=cps.scaled_pt_size*.5)

        ax.set_xlim(xr[0],xr[1])
        ax.set_ylim(yr[0],yr[1])
        ax.set_axis_off()
        ax.set_aspect('equal', adjustable='box')





#  Copyright (c) 2021-2025, Greg Michael
#  Licensed under BSD 3-Clause License. See LICENSE.txt for details.

__version__ = "3.6.1"

from .Chronologyfn import Chronologyfn
from .Productionfn import Productionfn
from .Cratercount import Cratercount
from .Craterplotset import Craterplotset
from .Craterplot import Craterplot
from .Craterpdf import Craterpdf
from .Epochs import Epochs
from .Spatialcount import Spatialcount
from .Randomnessanalysis import Randomnessanalysis

from .miscellaneous import *
from .constants import *
